@@ -1,15 +1,12 @@
 /*
    Raw Photo Parser
-   Copyright 2004-2005 by Dave Coffin, dcoffin a cybercom o net
+   Copyright 2004-2006 by Dave Coffin, dcoffin a cybercom o net
 
-   This program extracts thumbnail images (preferably JPEGs)
-   from any raw digital camera formats that have them, and
-   shows table contents.
+   This program displays raw metadata for all raw photo formats.
+   It is free for all uses.
 
-   This code is free for all uses.
-
-   $Revision: 1.55 $
-   $Date: 2005/11/24 01:29:21 $
+   $Revision: 1.62 $
+   $Date: 2006/08/18 02:51:53 $
  */
 
 #include <stdio.h>
@@ -39,15 +36,8 @@ typedef unsigned short ushort;
 
 FILE *ifp;
 short order;
-char *fname;
-char make[128], model[128], model2[128], thumb_head[128];
-int width, height, offset, length, bps, is_dng;
-int thumb_offset, thumb_length, thumb_layers;
-
-struct decode {
-  struct decode *branch[2];
-  int leaf;
-} first_decode[640], *free_decode;
+char *fname, make[128], model[128], model2[128];
+int is_dng;
 
 ushort sget2 (uchar *s)
 {
@@ -88,6 +78,17 @@ float int_to_float (int i)
   return u.f;
 }
 
+double get_double()
+{
+  union { char c[8]; double d; } u;
+  int i, rev;
+
+  rev = 7 * ((order == 0x4949) == (ntohs(0x1234) == 0x1234));
+  for (i=0; i < 8; i++)
+    u.c[i ^ rev] = fgetc(ifp);
+  return u.d;
+}
+
 void tiff_dump(int base, int tag, int type, int count, int level)
 {
   int save, j, num, den;
@@ -99,6 +100,7 @@ void tiff_dump(int base, int tag, int type, int count, int level)
   save = ftell(ifp);
   printf("%*stag=0x%x %d, type=%d, count=%d, offset=%06x, data=",
 	level*2, "", tag, tag, type, count, save);
+  if (tag == 34310) goto quit;
   if (type==2) putchar('\"');
   for (j = 0; j < count && j < DLEN; j++)
     switch (type) {
@@ -119,9 +121,11 @@ void tiff_dump(int base, int tag, int type, int count, int level)
 	num = get4();
 	den = get4();
 	printf (" %d/%d", num, den);
+//	printf (" %lf", (double) num/den);
 	break;
     }
   if (type==2) putchar('\"');
+quit:
   putchar('\n');
   fseek (ifp, save, SEEK_SET);
 }
@@ -194,7 +198,7 @@ void nikon_decrypt (uchar ci, uchar cj, int tag, int i, int size, uchar *buf)
 
 int parse_tiff_ifd (int base, int level);
 
-void nef_parse_makernote (base)
+void parse_makernote (base)
 {
   int offset=0, entries, tag, type, count, val, save;
   unsigned serial=0, key=0;
@@ -249,33 +253,14 @@ void nef_parse_makernote (base)
       fread (buf98, sizeof buf98, 1, ifp);
     if (tag == 0xa7)
       key = fgetc(ifp)^fgetc(ifp)^fgetc(ifp)^fgetc(ifp);
-
-    if (tag == 0x100 && type == 7 && !strncmp(make,"OLYMPUS",7)) {
-      thumb_offset = ftell(ifp);
-      thumb_length = count;
-    }
-    if (tag == 0x280 && type == 1) {	/* EPSON */
-      strcpy (thumb_head, "\xff");
-      thumb_offset = ftell(ifp)+1;
-      thumb_length = count-1;
-    }
-    if (strstr(make,"Minolta") || strstr(make,"MINOLTA")) {
-      switch (tag) {
-	case 0x81:
-	  thumb_offset = ftell(ifp);
-	  thumb_length = count;
-	  break;
-	case 0x88:
-	  thumb_offset = get4() + base;
-	  break;
-	case 0x89:
-	  thumb_length = get4();
-      }
-    }
     if (!strcmp (buf,"OLYMP") && tag >> 8 == 0x20)
       parse_tiff_ifd (base, 3);
     if (tag == 0xe01)
       parse_nikon_capture_note (count);
+    if (tag == 0xb028) {
+      fseek (ifp, get4(), SEEK_SET);
+      parse_tiff_ifd (base, 3);
+    }
     fseek (ifp, save+12, SEEK_SET);
   }
   nikon_decrypt (serial, key, 0x91, 4, sizeof buf91, buf91);
@@ -287,11 +272,11 @@ void nef_parse_makernote (base)
   order = sorder;
 }
 
-void nef_parse_exif(int base)
+void parse_exif(int base)
 {
   int entries, tag, type, count, save;
 
-  puts("Nikon EXIF tag:");
+  puts("EXIF table:");
   entries = get2();
   while (entries--) {
     save = ftell(ifp);
@@ -300,7 +285,7 @@ void nef_parse_exif(int base)
     count= get4();
     tiff_dump (base, tag, type, count, 1);
     if (tag == 0x927c)
-      nef_parse_makernote (base);
+      parse_makernote (base);
     fseek (ifp, save+12, SEEK_SET);
   }
 }
@@ -326,13 +311,12 @@ void sony_decrypt (unsigned *data, int len, int start, int key)
 
 int parse_tiff_ifd (int base, int level)
 {
-  int entries, tag, type, count, slen, save, save2, val, i;
-  int comp=0;
+  int entries, tag, type, count, slen, save, save2, i;
   unsigned *buf, sony_offset=0, sony_length=0, sony_key=0;
   FILE *sfp;
 
   entries = get2();
-  if (entries > 255) return 1;
+  if (entries > 1024) return 1;
   while (entries--) {
     save = ftell(ifp);
     tag  = get2();
@@ -340,44 +324,8 @@ int parse_tiff_ifd (int base, int level)
     count= get4();
     slen = count;
     if (slen > 128) slen = 128;
-
     tiff_dump (base, tag, type, count, level);
-
-    save2 = ftell(ifp);
-    if (type == 3)			/* short int */
-      val = get2();
-    else
-      val = get4();
-    fseek (ifp, save2, SEEK_SET);
-
-    if (tag > 50700 && tag < 50800)
-      is_dng = 1;
-
-    if (level == 3) {			/* Olympus E-1 and E-300 */
-      if (type == 4) {
-	if (tag == 0x101)
-	  thumb_offset = val;
-	else if (tag == 0x102)
-	  thumb_length = val;
-      }
-      goto cont;
-    }
     switch (tag) {
-      case 0x100:			/* ImageWidth */
-	if (!width)  width = val;
-	break;
-      case 0x101:			/* ImageHeight */
-	if (!height) height = val;
-	break;
-      case 0x102:			/* Bits per sample */
-	if (bps) break;
-	bps = val;
-	if (count == 1)
-	  thumb_layers = 1;
-	break;
-      case 0x103:			/* Compression */
-	comp = val;
-	break;
       case 0x10f:			/* Make tag */
 	fgets (make, slen, ifp);
 	break;
@@ -387,14 +335,6 @@ int parse_tiff_ifd (int base, int level)
       case 33405:			/* Model2 tag */
 	fgets (model2, slen, ifp);
 	break;
-      case 0x111:			/* StripOffset */
-	if (!offset || is_dng) offset = val;
-	break;
-      case 0x117:			/* StripByteCounts */
-	if (!length || is_dng) length = val;
-	if (offset > val && !strncmp(make,"KODAK",5) && !is_dng)
-	  offset -= val;
-	break;
       case 0x14a:			/* SubIFD tag */
 	save2 = ftell(ifp);
 	for (i=0; i < count; i++) {
@@ -404,25 +344,20 @@ int parse_tiff_ifd (int base, int level)
 	  parse_tiff_ifd (base, level+1);
 	}
 	break;
-      case 0x201:
-	if (!thumb_offset || (strncmp(make,"OLYMPUS",7)
-			   && strncmp(make,"SONY",4)))
-	  thumb_offset = val;
+      case 29184: sony_offset = get4();  break;
+      case 29185: sony_length = get4();  break;
+      case 29217: sony_key    = get4();  break;
+      case 33424:
+	puts("Kodak private data:");
+	fseek (ifp, get4()+base, SEEK_SET);
+	parse_tiff_ifd (base, level+1);
 	break;
-      case 0x202:
-	if (!thumb_length || (strncmp(make,"OLYMPUS",7)
-			   && strncmp(make,"SONY",4)))
-	  thumb_length = val;
-	break;
-      case 29184: sony_offset = val;  break;
-      case 29185: sony_length = val;  break;
-      case 29217: sony_key = get4();  break;
       case 34310:
 	parse_mos(0);
 	break;
       case 34665:
 	fseek (ifp, get4()+base, SEEK_SET);
-	nef_parse_exif (base);
+	parse_exif (base);
 	break;
       case 50706:
 	is_dng = 1;
@@ -433,7 +368,6 @@ int parse_tiff_ifd (int base, int level)
 	fseek (ifp, get4()+base, SEEK_SET);
 	parse_tiff_ifd (base, level+1);
     }
-cont:
     fseek (ifp, save+12, SEEK_SET);
   }
   if (sony_length && (buf = malloc(sony_length))) {
@@ -451,11 +385,6 @@ cont:
     ifp = sfp;
     free (buf);
   }
-  if (((comp == 6 && !strcmp(make,"Canon")) ||
-       (comp == 7 && is_dng)) && offset) {
-    thumb_offset = offset;
-    thumb_length = length;
-  }
   return 0;
 }
 
@@ -464,9 +393,8 @@ cont:
  */
 void parse_tiff (int base)
 {
-  int doff, spp=3, ifd=0;
+  int doff, ifd=0;
 
-  width = height = offset = length = bps = is_dng = 0;
   fseek (ifp, base, SEEK_SET);
   order = get2();
   if (order != 0x4949 && order != 0x4d4d) return;
@@ -476,25 +404,6 @@ void parse_tiff (int base)
     printf ("IFD #%d:\n", ifd++);
     if (parse_tiff_ifd (base, 0)) break;
   }
-  if (is_dng) return;
-
-  if (strncmp(make,"KODAK",5))
-    thumb_layers = 0;
-  if (!strncmp(make,"Kodak",5)) {
-    fseek (ifp, 12+base, SEEK_SET);
-    puts ("\nSpecial Kodak image directory:");
-    parse_tiff_ifd (base, 0);
-  }
-  if (!strncmp(model,"DCS460A",7)) {
-    spp = 1;
-    thumb_layers = 0;
-  }
-  if (!thumb_length && offset) {
-    thumb_offset = offset;
-    sprintf (thumb_head, "P%d %d %d %d\n",
-	spp > 1 ? 6:5, width, height, (1 << bps) - 1);
-    thumb_length = width * height * spp * ((bps+7)/8);
-  }
 }
 
 void parse_minolta()
@@ -514,9 +423,6 @@ void parse_minolta()
     }
     fseek (ifp, save+len+8, SEEK_SET);
   }
-  strcpy (thumb_head, "\xff");
-  thumb_offset++;
-  thumb_length--;
 }
 
 /*
@@ -526,6 +432,7 @@ void parse_ciff (int offset, int length, int level)
 {
   int tboff, nrecs, i, j, type, len, dlen, roff, aoff=0, save;
   char c, name[256];
+  ushort key[2];
 
   fseek (ifp, offset+length-4, SEEK_SET);
   tboff = get4() + offset;
@@ -573,8 +480,14 @@ void parse_ciff (int offset, int length, int level)
 	putchar('\"');
 	break;
       case 0x10:			/* word values */
+	key[0] = get2();
+	fseek (ifp, -2, SEEK_CUR);
+	if (type == 0x1032 && key[0] == 1040)
+	  key[1] = 17907;
+	else key[0] = key[1] = 0;
 	for (j = 0; j < dlen; j+=2)
-	  printf ("%c%5u",(j & 31) || dlen < 16 ? ' ':'\n', get2());
+	  printf ("%c%5u",(j & 31) || dlen < 16 ? ' ':'\n',
+		get2() ^ key[(j >> 1) & 1]);
 	break;
       case 0x18:			/* dword values */
 	for (j = 0; j < dlen; j+=4)
@@ -588,10 +501,6 @@ void parse_ciff (int offset, int length, int level)
       strcpy (make, name);
       strcpy (model, name + strlen(make)+1);
     }
-    if (type == 0x2007) {		/* Found the JPEG thumbnail */
-      thumb_offset = aoff;
-      thumb_length = len;
-    }
   }
 }
 
@@ -613,14 +522,6 @@ int parse_jpeg (int offset)
     parse_tiff (save+6);
     fseek (ifp, save+len, SEEK_SET);
   }
-  if (thumb_offset) {
-    fseek (ifp, thumb_offset, SEEK_SET);
-    for (hlen=0; hlen < 1024; hlen++) {
-      if (get2() == 0xffd8) return 1;
-      thumb_offset += 2;
-    }
-  }
-  thumb_length = 0;
   return 1;
 }
 
@@ -662,24 +563,22 @@ void parse_mos(int level)
 
   save = ftell(ifp);
   while (1) {
-    fread (data, 1, 8, ifp);
-    if (strcmp(data,"PKTS")) break;
-    strcpy (model, "Valeo");
-    printf ("%*s%s ", level, "", data);
+    if (get4() != 0x504b5453) break;
+    get4();
+    printf ("%*sPKTS ", level, "");
     fread (data, 1, 40, ifp);
     skip = get4();
     printf ("%s %d bytes: ", data, skip);
+    if (!strcmp(data,"pattern_ratation_angle")) {
+      printf ("%d\n", get2());
+      continue;
+    }
     if (!strcmp(data,"icc_camera_to_tone_matrix")) {
       for (i=0; i < skip/4; i++)
 	printf ("%f ", int_to_float(get4()));
       putchar('\n');
       continue;
     }
-    if (!strcmp(data,"JPEG_preview_data")) {
-      thumb_head[0] = 0;
-      thumb_offset = ftell(ifp);
-      thumb_length = skip;
-    }
     fread (data, 1, sizeof data, ifp);
     fseek (ifp, -sizeof data, SEEK_CUR);
     data[sizeof data - 1] = 0;
@@ -704,33 +603,9 @@ void parse_rollei()
       *val++ = 0;
     else
       val = line + strlen(line);
-    if (!strcmp(line,"HDR"))
-      thumb_offset = atoi(val);
-    if (!strcmp(line,"TX "))
-      width = atoi(val);
-    if (!strcmp(line,"TY "))
-      height = atoi(val);
   } while (strncmp(line,"EOHD",4));
   strcpy (make, "Rollei");
   strcpy (model, "d530flex");
-  thumb_length = width*height*2;
-}
-
-void rollei_decode (FILE *tfp)
-{
-  ushort data;
-  int row, col;
-
-  fseek (ifp, thumb_offset, SEEK_SET);
-  fprintf (tfp, "P6\n%d %d\n255\n", width, height);
-  for (row=0; row < height; row++)
-    for (col=0; col < width; col++) {
-      fread (&data, 2, 1, ifp);
-      data = ntohs(data);
-      putc (data << 3, tfp);
-      putc (data >> 5  << 2, tfp);
-      putc (data >> 11 << 3, tfp);
-    }
 }
 
 void get_utf8 (int offset, char *buf, int len)
@@ -756,7 +631,7 @@ void get_utf8 (int offset, char *buf, int len)
 
 void parse_foveon()
 {
-  int entries, img=0, off, len, tag, save, i, j, k, pent, poff[256][2];
+  int entries, off, len, tag, save, i, j, k, pent, poff[256][2];
   char name[128], value[128], camf[0x20000], *pos, *cp, *dp;
   unsigned val, key, type, num, ndim, dim[3];
 
@@ -791,15 +666,8 @@ void parse_foveon()
 	printf ("columns %4d, "	, get4());
 	printf ("rows %4d, "	, get4());
 	printf ("rowsize %d\n"	, get4());
-	if (parse_jpeg (off+28)) {
-	  thumb_offset = off+28;
-	  thumb_length = len-28;
-	}
+	parse_jpeg (off+28);
 	order = 0x4949;
-	if (++img == 2 && !thumb_length) {
-	  thumb_offset = off;
-	  thumb_length = 1;
-	}
 	break;
       case 0x464d4143:			/* CAMF */
 	printf ("type %d, ", get4());
@@ -909,143 +777,6 @@ next:
   }
 }
 
-void foveon_tree (unsigned huff[1024], unsigned code)
-{
-  struct decode *cur;
-  int i, len;
-
-  cur = free_decode++;
-  if (code) {
-    for (i=0; i < 1024; i++)
-      if (huff[i] == code) {
-	cur->leaf = i;
-	return;
-      }
-  }
-  if ((len = code >> 27) > 26) return;
-  code = (len+1) << 27 | (code & 0x3ffffff) << 1;
-
-  cur->branch[0] = free_decode;
-  foveon_tree (huff, code);
-  cur->branch[1] = free_decode;
-  foveon_tree (huff, code+1);
-}
-
-void foveon_decode (FILE *tfp)
-{
-  int bwide, row, col, bit=-1, c, i;
-  char *buf;
-  struct decode *dindex;
-  short pred[3];
-  unsigned huff[1024], bitbuf=0;
-
-  fseek (ifp, thumb_offset+16, SEEK_SET);
-  width  = get4();
-  height = get4();
-  bwide  = get4();
-  fprintf (tfp, "P6\n%d %d\n255\n", width, height);
-  if (bwide > 0) {
-    buf = malloc(bwide);
-    for (row=0; row < height; row++) {
-      fread  (buf, 1, bwide, ifp);
-      fwrite (buf, 3, width, tfp);
-    }
-    free (buf);
-    return;
-  }
-  for (i=0; i < 256; i++)
-    huff[i] = get4();
-  memset (first_decode, 0, sizeof first_decode);
-  free_decode = first_decode;
-  foveon_tree (huff, 0);
-
-  for (row=0; row < height; row++) {
-    memset (pred, 0, sizeof pred);
-    if (!bit) get4();
-    for (col=bit=0; col < width; col++) {
-      for (c=0; c < 3; c++) {
-	for (dindex=first_decode; dindex->branch[0]; ) {
-	  if ((bit = (bit-1) & 31) == 31)
-	    for (i=0; i < 4; i++)
-	      bitbuf = (bitbuf << 8) + fgetc(ifp);
-	  dindex = dindex->branch[bitbuf >> bit & 1];
-	}
-	pred[c] += dindex->leaf;
-	fputc (pred[c], tfp);
-      }
-    }
-  }
-}
-
-void kodak_yuv_decode (FILE *tfp)
-{
-  uchar c, blen[384];
-  unsigned row, col, len, bits=0;
-  INT64 bitbuf=0;
-  int i, li=0, si, diff, six[6], y[4], cb=0, cr=0, rgb[3];
-  ushort *out, *op;
-
-  fseek (ifp, thumb_offset, SEEK_SET);
-  width = (width+1) & -2;
-  height = (height+1) & -2;
-  fprintf (tfp, "P6\n%d %d\n65535\n", width, height);
-  out = malloc (width * 12);
-  if (!out) {
-    fprintf (stderr, "kodak_yuv_decode() malloc failed!\n");
-    exit(1);
-  }
-
-  for (row=0; row < height; row+=2) {
-    for (col=0; col < width; col+=2) {
-      if ((col & 127) == 0) {
-	len = (width - col + 1) * 3 & -4;
-	if (len > 384) len = 384;
-	for (i=0; i < len; ) {
-	  c = fgetc(ifp);
-	  blen[i++] = c & 15;
-	  blen[i++] = c >> 4;
-	}
-	li = bitbuf = bits = y[1] = y[3] = cb = cr = 0;
-	if (len % 8 == 4) {
-	  bitbuf  = fgetc(ifp) << 8;
-	  bitbuf += fgetc(ifp);
-	  bits = 16;
-	}
-      }
-      for (si=0; si < 6; si++) {
-	len = blen[li++];
-	if (bits < len) {
-	  for (i=0; i < 32; i+=8)
-	    bitbuf += (INT64) fgetc(ifp) << (bits+(i^8));
-	  bits += 32;
-	}
-	diff = bitbuf & (0xffff >> (16-len));
-	bitbuf >>= len;
-	bits -= len;
-	if ((diff & (1 << (len-1))) == 0)
-	  diff -= (1 << len) - 1;
-	six[si] = diff;
-      }
-      y[0] = six[0] + y[1];
-      y[1] = six[1] + y[0];
-      y[2] = six[2] + y[3];
-      y[3] = six[3] + y[2];
-      cb  += six[4];
-      cr  += six[5];
-      for (i=0; i < 4; i++) {
-	op = out + ((i >> 1)*width + col+(i & 1)) * 3;
-	rgb[0] = y[i] + 1.40200/2 * cr;
-	rgb[1] = y[i] - 0.34414/2 * cb - 0.71414/2 * cr;
-	rgb[2] = y[i] + 1.77200/2 * cb;
-	for (c=0; c < 3; c++)
-	  if (rgb[c] > 0) op[c] = htons(rgb[c]);
-      }
-    }
-    fwrite (out, sizeof *out, width*6, tfp);
-  }
-  free(out);
-}
-
 void parse_fuji (int offset)
 {
   int entries, tag, len;
@@ -1069,6 +800,7 @@ void parse_fuji (int offset)
 void parse_phase_one (int base)
 {
   unsigned entries, tag, type, len, data, save;
+  unsigned meta=0, wide=0, high=0, i, j;
   char str[256];
 
   fseek (ifp, base, SEEK_SET);
@@ -1083,8 +815,12 @@ void parse_phase_one (int base)
     len  = get4();
     data = get4();
     save = ftell(ifp);
-    printf ("Phase One tag=0x%x, type=%d, len=%2d, data = 0x%x\n",
+    printf ("Phase One tag=0x%x, type=%d, len=%2d, data = 0x%x",
 		tag, type, len, data);
+    if (type == 4 && len == 4 && data > 0xfffffff)
+      printf (" = %f", int_to_float(data));
+    putchar ('\n');
+    if (tag == 0x110) meta = base+data;
     if (len > 4)
       fseek (ifp, base+data, SEEK_SET);
     if (type == 1 && len < 256) {
@@ -1096,14 +832,78 @@ void parse_phase_one (int base)
 	printf ("%f ", int_to_float(get4()));
       puts ("");
     }
-    if (tag == 0x110) {
-      thumb_offset = data + base;
-      thumb_length = len;
-    }
     fseek (ifp, save, SEEK_SET);
   }
   strcpy (make, "Phase One");
   strcpy (model, "unknown");
+  if (!meta) return;
+  fseek (ifp, meta, SEEK_SET);
+  order = get2();
+  fseek (ifp, 6, SEEK_CUR);
+  fseek (ifp, meta+get4(), SEEK_SET);
+  entries = get4();
+  get4();
+  while (entries--) {
+    tag  = get4();
+    len  = get4();
+    data = get4();
+    save = ftell(ifp);
+    printf ("Phase One meta tag=0x%x, len=%2d, offset = 0x%x, data = ",
+		tag, len, data);
+    if (!((0x000801f4 >> (tag-0x400)) & 1)) putchar ('\n');
+    fseek (ifp, meta+data, SEEK_SET);
+    switch (tag) {
+      case 0x400:
+	for (i=0; i < len; i+=2)
+	  printf ("%5u%c", get2(), (i & 6) == 6 || i == len-2 ? '\n':' ');
+	break;
+      case 0x401:
+	for (i=0; i < 16; i+=2)
+	  printf ("%6u%c", get2(), (i & 14) == 14 || i == len-2 ? '\n':' ');
+	for (; i < len; i+=4)
+	  printf ("%9.6f%c", int_to_float(get4()),
+		(i & 28) == 12 || i == len-4 ? '\n':' ');
+	break;
+      case 0x402:
+	printf ("%f\n", int_to_float (data));
+	break;
+      case 0x404: case 0x405: case 0x406: case 0x407:
+	fread (str, 256, 1, ifp);
+	puts (str);
+	break;
+      case 0x408: case 0x413:
+	printf ("%lf\n", get_double());
+	break;
+      case 0x40b: case 0x410: case 0x416:
+	for (i=0; i < len; i+=2)
+	  printf ("%6u%c", get2(), (i & 14) == 14 || i == len-2 ? '\n':' ');
+	break;
+      case 0x40f: case 0x418: case 0x419: case 0x41a:
+	for (i=0; i < 4; i++)
+	  printf ("%02X%c", fgetc(ifp), i == 3 ? '\n':' ');
+	for (; i < len; i+=4)
+	  printf ("%e%c", int_to_float(get4()), i == len-4 ? '\n':' ');
+	break;
+      case 0x412:
+	for (i=0; i < 36; i+=4) {
+	  printf ("%u ", j=get4());
+	  if (i ==  4) wide = j;
+	  if (i == 12) high = j*2;
+	}
+	printf ("%u\n", get2());
+	for (i=0; i < wide*high; i++)
+	  printf ("%9.6f%c", int_to_float(get4()),
+			i % wide == wide-1 ? '\n':' ');
+	for (i=0; i < wide*high; i++)
+	  printf ("%5u%c", get2(), i % wide == wide-1 ? '\n':' ');
+	break;
+      default:
+	for (i=0; i < len; i++)
+	  printf ("%02X%c", fgetc(ifp),
+		(i & 15) == 15 || i == len-1 ? '\n':' ');
+    }
+    fseek (ifp, save, SEEK_SET);
+  }
 }
 
 char *memmem (char *haystack, size_t haystacklen,
@@ -1120,14 +920,12 @@ char *memmem (char *haystack, size_t haystacklen,
    Identify which camera created this file, and set global variables
    accordingly.	 Return nonzero if the file cannot be decoded.
  */
-int identify()
+void identify()
 {
-  char head[32], thumb_name[256], *thumb, *rgb, *cp;
-  unsigned hlen, fsize, toff, tlen, lsize, i;
-  FILE *tfp;
+  char head[32], *cp;
+  unsigned hlen, fsize, toff, tlen;
 
   make[0] = model[0] = model2[0] = is_dng = 0;
-  thumb_head[0] = thumb_offset = thumb_length = thumb_layers = 0;
   order = get2();
   hlen = get4();
   fseek (ifp, 0, SEEK_SET);
@@ -1144,6 +942,8 @@ int identify()
       fseek (ifp, hlen, SEEK_SET);
     } else
       parse_tiff (0);
+  } else if (!memcmp (head,"NDF0",4)) {
+    parse_tiff (12);
   } else if (!memcmp (head,"\0MRM",4)) {
     parse_minolta();
   } else if (!memcmp (head,"FUJIFILM",8)) {
@@ -1153,8 +953,6 @@ int identify()
     parse_fuji (92);
     if (toff > 120) parse_fuji (120);
     parse_tiff (toff+12);
-    thumb_offset = toff;
-    thumb_length = tlen;
   } else if (!memcmp (head,"RIFF",4)) {
     fseek (ifp, 0, SEEK_SET);
     parse_riff(0);
@@ -1162,72 +960,7 @@ int identify()
     parse_rollei();
   else if (!memcmp (head,"FOVb",4))
     parse_foveon();
-  fseek (ifp, 8, SEEK_SET);
-  parse_mos(0);
-  fseek (ifp, 3472, SEEK_SET);
-  parse_mos(0);
   parse_jpeg(0);
-  if (model[0] == 0) {
-    fprintf (stderr, "%s: unsupported file format.\n", fname);
-    return 1;
-  }
-  fprintf (stderr, "Findings for %s:\n", fname);
-  fprintf (stderr, "Make   is \"%s\"\n", make);
-  fprintf (stderr, "Model  is \"%s\"\n", model);
-  if (model2[0])
-    fprintf (stderr, "Model2 is \"%s\"\n", model2);
-
-  if (!thumb_length) {
-    fprintf (stderr, "Thumbnail image not found\n");
-    return 0;
-  }
-  strcpy (thumb_name, fname);
-  strcat (thumb_name, ".thumb");
-  tfp = fopen (thumb_name, "wb");
-  if (!tfp) {
-    perror(thumb_name);
-    return 0;
-  }
-  if (is_dng) goto dng_skip;
-  if (!strncmp(model,"DCS Pro",7)) {
-    kodak_yuv_decode (tfp);
-    goto done;
-  }
-  if (!strcmp(make,"Rollei")) {
-    rollei_decode (tfp);
-    goto done;
-  }
-  if (!strcmp(make,"SIGMA")) {
-    foveon_decode (tfp);
-    goto done;
-  }
-dng_skip:
-  thumb = (char *) malloc(thumb_length);
-  if (!thumb) {
-    fprintf (stderr, "Cannot allocate %d bytes!!\n", thumb_length);
-    exit(1);
-  }
-  fseek (ifp, thumb_offset, SEEK_SET);
-  fread (thumb, 1, thumb_length, ifp);
-  if (thumb_layers && !is_dng) {
-    rgb = (char *) malloc(thumb_length);
-    if (!rgb) {
-      fprintf (stderr, "Cannot allocate %d bytes!!\n", thumb_length);
-      exit(1);
-    }
-    lsize = thumb_length/3;
-    for (i=0; i < thumb_length; i++)
-      rgb[(i%lsize)*3 + i/lsize] = thumb[i];
-    free(thumb);
-    thumb = rgb;
-  }
-  fputs (thumb_head, tfp);
-  fwrite(thumb, 1, thumb_length, tfp);
-  free (thumb);
-done:
-  fclose (tfp);
-  fprintf (stderr, "Thumbnail image written to %s.\n", thumb_name);
-  return 0;
 }
 
 int main(int argc, char **argv)
@@ -1237,12 +970,11 @@ int main(int argc, char **argv)
   if (argc == 1)
   {
     fprintf (stderr,
-    "\nRaw Photo Parser and Thumbnail Extracter"
+    "\nRaw Photo Parser"
     "\nby Dave Coffin, dcoffin a cybercom o net"
-    "\n\nUsage:  %s [options] file1.crw file2.crw ...\n", argv[0]);
+    "\n\nUsage:  %s file1.crw file2.crw ...\n", argv[0]);
     return 1;
   }
-
   for (arg=1; arg < argc; arg++)
   {
     fname = argv[arg];
